import json
from typing import List, Optional, Type
from uuid import uuid4

from jinja2 import Environment, PackageLoader

from unfolded.map_sdk._version import __version__
from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.transport.base import BaseNonInteractiveTransport
from unfolded.map_sdk.types import ResponseClass
from unfolded.map_sdk.utils.serialization import serialize_action


class HTMLTransport(BaseNonInteractiveTransport):
    """Transport used in a static HTML map"""

    action_list: List[Action]

    def __init__(self) -> None:
        super().__init__()
        self.action_list = []

    def send_action(
        self,
        *,
        action: Action,
        response_class: Optional[Type[ResponseClass]] = None,
    ) -> None:
        # pylint:disable=unused-argument
        self.action_list.append(action)

    @property
    def serialized_actions(self):
        return [
            serialize_action(action, renderer="html")[0] for action in self.action_list
        ]

    def render_template(self, style, basemaps) -> str:
        self.rendered = True
        j2_env = Environment(
            loader=PackageLoader("unfolded.map_sdk"),
            autoescape=True,
            trim_blocks=True,
            keep_trailing_newline=True,
        )
        template = j2_env.get_template("html_map_sdk.j2")

        html_str = template.render(
            version=__version__,
            actions=json.dumps(self.serialized_actions),
            div_id=uuid4(),
            style=json.dumps(style),
            basemaps=json.dumps(basemaps),
        )
        return html_str



from typing import List, Optional

from unfolded.map_sdk.api.base import CamelCaseBaseModel
from unfolded.map_sdk.api.dataset_api import (
    DatasetApiInteractiveMixin,
    DatasetApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.event_api import EventApiInteractiveMixin
from unfolded.map_sdk.api.filter_api import (
    FilterApiInteractiveMixin,
    FilterApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.layer_api import (
    LayerApiInteractiveMixin,
    LayerApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.map_api import (
    MapApiInteractiveMixin,
    MapApiNonInteractiveMixin,
    MapStyleCreationProps,
)
from unfolded.map_sdk.environment import default_height
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
    BaseTransport,
)


class BasemapParams(CamelCaseBaseModel):
    custom_map_styles: Optional[List[MapStyleCreationProps]]
    initial_map_style_id: Optional[str]


DEFAULT_MAP_STYLE = {
    "height": default_height(),
    "width": "100%",
}


class BaseMap:
    """
    Base class for all map types (both widget and non-widget)
    """

    transport: BaseTransport


class BaseInteractiveMap(
    BaseMap,
    MapApiInteractiveMixin,
    DatasetApiInteractiveMixin,
    FilterApiInteractiveMixin,
    LayerApiInteractiveMixin,
    EventApiInteractiveMixin,
):
    transport: BaseInteractiveTransport
    pass


class BaseNonInteractiveMap(
    BaseMap,
    MapApiNonInteractiveMixin,
    DatasetApiNonInteractiveMixin,
    FilterApiNonInteractiveMixin,
    LayerApiNonInteractiveMixin,
):
    transport: BaseNonInteractiveTransport
    pass



import importlib
import inspect
import logging

import pytest
from pydantic import parse_obj_as

from unfolded.map_sdk.api.dataset_api import (
    Dataset,
    LocalDataset,
    RasterTileDataset,
    RasterTileLocalItemMetadata,
    VectorTileDataset,
    VectorTileLocalMetadata,
)
from unfolded.map_sdk.utils.serialization import serialize_action

from .fixtures import SERIALIZED_ACTIONS
from .fixtures.dataset_api import (
    LOCAL_RESPONSE,
    RASTER_TILE_RESPONSE,
    VECTOR_TILE_RESPONSE,
)


class TestActionSerialization:
    def test_ensure_all_actions_tested(self):
        """All Actions should be listed in SERIALIZED_ACTIONS"""

        MAP_API = importlib.import_module("unfolded.map_sdk.api.map_api")
        DATASET_API = importlib.import_module("unfolded.map_sdk.api.dataset_api")
        LAYER_API = importlib.import_module("unfolded.map_sdk.api.layer_api")
        FILTER_API = importlib.import_module("unfolded.map_sdk.api.filter_api")
        API_MODULES = [MAP_API, DATASET_API, LAYER_API, FILTER_API]

        all_actions = set()

        for api_module in API_MODULES:
            for name, obj in inspect.getmembers(api_module, inspect.isclass):
                if name.endswith("Action") and inspect.getmodule(obj) in API_MODULES:
                    all_actions.add(obj)

        listed_actions = set()
        for (action, _) in SERIALIZED_ACTIONS:
            listed_actions.add(action.__class__)

        diff_actions = all_actions.symmetric_difference(listed_actions)
        assert not diff_actions, f"Actions listed but not tested: {diff_actions}"

    @pytest.mark.parametrize("action,serialized", SERIALIZED_ACTIONS)
    def test_serialize_action(self, action, serialized, caplog):
        """All fields other than `type` and `messageId` should be included within the `data` key."""
        with caplog.at_level(logging.DEBUG):
            d, _ = serialize_action(action)

        # make sure there are no debug warnings in message serialization
        assert not caplog.text, f"Action not serialized properly: {caplog.text}"

        # messageId is a unique uuid and won't match
        d.pop("messageId")
        serialized.pop("messageId")

        assert (
            d == serialized
        ), f"Message not serialized as expected: {d} vs {serialized}"


class TestDataset:
    """Tests relating to dataset serialization/deserialization"""

    def test_deserialize(self):

        local_dataset = parse_obj_as(Dataset, LOCAL_RESPONSE)
        assert isinstance(local_dataset, LocalDataset)

        vector_tile_dataset = parse_obj_as(Dataset, VECTOR_TILE_RESPONSE)
        assert isinstance(vector_tile_dataset, VectorTileDataset)
        assert isinstance(vector_tile_dataset.metadata, VectorTileLocalMetadata)

        raster_tile_dataset = parse_obj_as(Dataset, RASTER_TILE_RESPONSE)
        assert isinstance(raster_tile_dataset, RasterTileDataset)
        assert isinstance(raster_tile_dataset.metadata, RasterTileLocalItemMetadata)



import re
import sys
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import AnyHttpUrl
from pydantic import Field as PydanticField
from pydantic import StrictStr, root_validator, validator

from unfolded.map_sdk.api.base import Action, CamelCaseBaseModel, Number
from unfolded.map_sdk.api.enums import (
    ActionType,
    BasicFieldType,
    DatasetType,
    FieldType,
    TimestampFieldType,
    VectorTileAttributeType,
)
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
)
from unfolded.map_sdk.utils.validators import validate_kwargs

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import geopandas as gpd
except ImportError:
    gpd = None


RGBColor = Tuple[Number, Number, Number]
"""Red, green and blue channels of a color in [0-255] range."""


class BaseField(CamelCaseBaseModel):

    name: StrictStr
    """Unique identifier of the field."""

    label: StrictStr
    """User facing field label."""

    type: FieldType
    """Type of the field."""


class BasicField(BaseField):
    """Contains semantics around a single field in a dataset table."""

    type: BasicFieldType


class TimestampField(BaseField):
    """Contains semantics around a single timestamp field in a dataset table."""

    type: TimestampFieldType = TimestampFieldType.TIMESTAMP
    time_format: StrictStr
    """moment.js time format of this field data."""


Field = Union[BasicField, TimestampField]
"""Union of available field types."""


class BaseDataset(CamelCaseBaseModel):

    id: StrictStr
    """Unique identifier of the dataset."""

    type: DatasetType
    """Type of dataset."""

    label: StrictStr
    """Displayable dataset label."""

    color: RGBColor
    """Color label of the dataset."""

    fields: List[Field]
    """Schema describring the fields of the dataset."""


class LocalDataset(BaseDataset):
    """Dataset record representing a local dataset, with data provided by the caller."""

    type = DatasetType.LOCAL


class DatasetWithData(LocalDataset):

    data: List[List[Any]] = PydanticField(..., repr=False)
    """Tabular data backing the dataset, where the order of fields matches the order in `Dataset.fields`."""


class VectorTileBaseMetadata(CamelCaseBaseModel):
    """Vector tileset base metadata."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    data_url: AnyHttpUrl
    """URL template for tiles, with {x}/{y}/{z} placeholders"""


class _VectorAttributes(CamelCaseBaseModel):
    """Attributes of a Vector layer."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    attribute: StrictStr
    """The name of this attribute."""

    type: VectorTileAttributeType
    """The type of this attribute's values."""

    min: Optional[Number]
    """The minimum value of this attribute (if numeric)."""

    max: Optional[Number]
    """The maximum value of this attribute (if numeric)."""


class _VectorLayer(CamelCaseBaseModel):
    """Layer attribute of the Vector TileJSON."""

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    attributes: List[_VectorAttributes]
    """Attributes in the layer."""


class _VectorTilestats(CamelCaseBaseModel):
    """Tilestats attribute of the Vector TileJSON."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(num_layers={len(self.layers)})"

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    layers: List[_VectorLayer]
    """Layers in the tileset."""


class VectorTileEmbeddedMetadata(CamelCaseBaseModel):
    """Minimal shape for the embedded JSON metadata in a Tippecanoe vector
    tileset metadata file.
    """

    class Config(CamelCaseBaseModel.Config):
        extra = "ignore"

    tilestats: _VectorTilestats
    """The tilestats for this tileset."""


class VectorTileLocalMetadata(VectorTileBaseMetadata):
    """Vector tileset metadata, following the metadata format generated
    by Tippecanoe. This metadata shape can be passed to the
    map to synchronously add a vector tileset.
    see https://github.com/mapbox/tippecanoe
    """

    # JSON metadata has multiple names in different contexts
    # maxzoom and minzoom are returned by studio in camelCase
    @root_validator(pre=True)
    # pylint:disable=no-self-argument
    def allow_alternate_aliases(cls, values):
        if not values.get("json"):
            values["json"] = values.get("metaJson")

        if not values.get("maxzoom"):
            values["maxzoom"] = values.get("maxZoom")

        if not values.get("minzoom"):
            values["minzoom"] = values.get("minZoom")

        return values

    class Config(VectorTileBaseMetadata.Config):
        extra = "ignore"

    metadata_url: Optional[AnyHttpUrl]
    """URL for tileset metadata."""

    bounds: Optional[Union[StrictStr, List[Number]]]
    """Tileset bounds, as an array or comma-delimited string in format "w,s,e,n"."""

    center: Optional[Union[StrictStr, List[Number]]]
    """Tileset center, as an array or comma-delimited string in format "lng,lat"."""

    maxzoom: Optional[Number]
    """Maximum zoom supported by the tileset."""

    minzoom: Optional[Number]
    """Minimum zoom supported by the tileset."""

    _json: Union[StrictStr, VectorTileEmbeddedMetadata, None] = PydanticField(
        ..., alias="json"
    )
    """Metadata for the tileset, as a JSON string or object."""


class VectorTileRemoteMetadata(VectorTileBaseMetadata):
    """Vector tileset metadata with a remote metadata URL. The remote metadata file
    should follow the metadata format generated by Tippecanoe. This metadata can
    be passed to the map to asynchronously load a vector tileset.
    see https://github.com/mapbox/tippecanoe
    """

    metadata_url: AnyHttpUrl
    """URL for tileset metadata."""


VectorTileMetadata = Union[VectorTileLocalMetadata, VectorTileRemoteMetadata]


class VectorTileDataset(BaseDataset):
    """Dataset record representing a vector tileset."""

    type = DatasetType.VECTOR_TILE

    metadata: VectorTileMetadata
    """Metadata about the vector tileset."""


class RasterTileBaseMetadata(CamelCaseBaseModel):
    """Raster tileset base metadata."""

    class Config(CamelCaseBaseModel.Config):
        extra = "allow"


class RasterTileLocalItemMetadata(RasterTileBaseMetadata):
    """Raster tileset metadata in STAC Item format. STAC version must be >= 1.0.0,
    and the EO and Raster STAC extensions are required. This metadata shape can
    be passed to the map to synchronously add a raster tileset.
    see https://github.com/radiantearth/stac-spec/blob/master/item-spec/item-spec.md
    """

    # pylint:disable=no-self-argument

    type: Literal["Feature"]
    """Type of the raster tileset."""

    stac_version: StrictStr = PydanticField(..., alias="stac_version")
    """STAC version for tileset."""

    stac_extensions: List[StrictStr] = PydanticField(..., alias="stac_extensions")
    """A list of extensions the Item implements."""

    assets: Dict[StrictStr, Any]
    """Dictionary of asset objects that can be downloaded, each with a unique key."""

    @validator("stac_version")
    def validate_stac_version(cls, v: str):
        if v[0] == "0":
            raise ValueError("stac_version must be 1.0 or higher")

        return v

    @validator("stac_extensions")
    def validate_stac_extensions(cls, v: List[str]):
        extension_regex = (
            r"https://stac-extensions.github.io/{ext}/v1\.\d\.\d/schema\.json"
        )
        if not any(re.match(extension_regex.format(ext="eo"), s) for s in v):
            raise ValueError("STAC eo extension, 1.0 or higher, is required")

        if not any(re.match(extension_regex.format(ext="raster"), s) for s in v):
            raise ValueError("STAC raster extension, 1.0 or higher, is required")

        return v


class RasterTileLocalCollectionMetadata(RasterTileBaseMetadata):
    # pylint:disable=no-self-argument

    type: Literal["Collection"]

    stac_version: StrictStr = PydanticField(..., alias="stac_version")
    """STAC version for tileset."""

    stac_extensions: List[StrictStr] = PydanticField(..., alias="stac_extensions")
    """A list of extensions the Item implements."""

    item_assets: Dict[StrictStr, Any] = PydanticField(..., alias="item_assets")
    """Dictionary of asset objects that items in the collection expose, each with a unique key."""

    @validator("stac_version")
    def validate_stac_version(cls, v: str):
        if v[0] == "0":
            raise ValueError("stac_version must be 1.0 or higher")

        return v

    @validator("stac_extensions")
    def validate_stac_extensions(cls, v: List[str]):
        extension_regex = (
            r"https://stac-extensions.github.io/{ext}/v1\.\d\.\d/schema\.json"
        )
        if not any(re.match(extension_regex.format(ext="eo"), s) for s in v):
            raise ValueError("STAC eo extension, 1.0 or higher, is required")

        if not any(re.match(extension_regex.format(ext="raster"), s) for s in v):
            raise ValueError("STAC raster extension, 1.0 or higher, is required")

        if not any(re.match(extension_regex.format(ext="item-assets"), s) for s in v):
            raise ValueError("STAC item-assets extension, 1.0 or higher, is required")

        return v


RasterTileLocalMetadata = Union[
    RasterTileLocalItemMetadata, RasterTileLocalCollectionMetadata
]


class RasterTileRemoteMetadata(RasterTileBaseMetadata):
    """Raster tileset metadata with a remote metadata URL. This metadata can
    be passed to the map to asynchronously load a raster tileset.
    """

    metadata_url: AnyHttpUrl
    """URL for tileset metadata."""


RasterTileMetadata = Union[RasterTileLocalMetadata, RasterTileRemoteMetadata]


class RasterTileDataset(BaseDataset):
    """Dataset record representing a raster tileset."""

    type = DatasetType.RASTER_TILE

    metadata: RasterTileMetadata
    """Metadata about the raster tileset."""


Dataset = Union[LocalDataset, VectorTileDataset, RasterTileDataset]


class _BaseDatasetCreationProps(CamelCaseBaseModel):

    id: Optional[StrictStr]
    """Unique identifier of the dataset."""

    type: DatasetType
    """Type of dataset."""

    label: Optional[StrictStr]
    """Displayable dataset label."""

    color: Optional[RGBColor]
    """Color label of the dataset."""


class _LocalDatasetCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a dataset with inline data."""

    type: DatasetType = PydanticField(default=DatasetType.LOCAL)

    data: Union[StrictStr, dict, List[List[Any]]]
    """Data used to create a dataset, in CSV, JSON, GeoJSON format."""


class _VectorTileDatasetCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a vector tile dataset with provided metadata."""

    type: DatasetType = DatasetType.VECTOR_TILE

    metadata: VectorTileLocalMetadata
    """Metadata about the vector tileset."""


class _VectorTileDatasetRemoteCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a vector tile dataset with remote metadata."""

    type: DatasetType = DatasetType.VECTOR_TILE

    metadata: VectorTileRemoteMetadata
    """Metadata about the vector tileset."""


class _RasterTileDatasetCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a raster tile dataset with provided metadata."""

    type: DatasetType = DatasetType.RASTER_TILE

    metadata: RasterTileLocalMetadata
    """Metadata about the raster tileset."""


class _RasterTileDatasetRemoteCreationProps(_BaseDatasetCreationProps):
    """A set of properties used to create a raster tile dataset with remote metadata."""

    type: DatasetType = DatasetType.RASTER_TILE

    metadata: RasterTileRemoteMetadata
    """Metadata about the raster tileset."""


_DatasetCreationProps = Union[
    _LocalDatasetCreationProps,
    _VectorTileDatasetCreationProps,
    _RasterTileDatasetCreationProps,
]


_TileDatasetCreationProps = Union[
    _VectorTileDatasetRemoteCreationProps, _RasterTileDatasetRemoteCreationProps
]


class _DatasetUpdateProps(CamelCaseBaseModel):
    """A set of properties used to update a dataset."""

    label: Optional[StrictStr]
    """Displayable dataset label."""

    color: Optional[RGBColor]
    """Color label of the dataset."""


###########
# ACTIONS #
###########


class GetDatasetsAction(Action):
    """Action payload sent with `get_datasets` calls"""

    type = ActionType.GET_DATASETS


class GetDatasetByIdAction(Action):
    """Action payload sent with `get_dataset_by_id` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id"]

    type = ActionType.GET_DATASET_BY_ID
    dataset_id: StrictStr


class AddDatasetAction(Action):
    """Action payload sent with `add_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset"]
        options = ["auto_create_layers", "center_map"]

    type = ActionType.ADD_DATASET
    dataset: _DatasetCreationProps
    auto_create_layers: bool
    center_map: bool


class AddTileDatasetAction(Action):
    """Action payload sent with `add_tile_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset"]
        options = ["auto_create_layers", "center_map"]

    type = ActionType.ADD_TILE_DATASET
    dataset: _TileDatasetCreationProps
    auto_create_layers: bool
    center_map: bool


class UpdateDatasetAction(Action):
    """Action payload sent with `update_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id", "values"]

    type = ActionType.UPDATE_DATASET
    dataset_id: StrictStr
    values: _DatasetUpdateProps


class RemoveDatasetAction(Action):
    """Action payload sent with `remove_dataset` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id"]

    type = ActionType.REMOVE_DATASET
    dataset_id: StrictStr


class ReplaceDatasetAction(Action):
    """Action payload sent with `replace_dataset` calls"""

    class Meta(Action.Meta):
        args = ["this_dataset_id", "with_dataset"]

    type = ActionType.REPLACE_DATASET
    this_dataset_id: StrictStr
    with_dataset: _DatasetCreationProps


class GetDatasetWithDataAction(Action):
    """Action payload sent with `get_dataset_with_data` calls"""

    class Meta(Action.Meta):
        args = ["dataset_id"]

    type = ActionType.GET_DATASET_WITH_DATA
    dataset_id: StrictStr


#########
# UTILS #
#########


def normalize_dataset(
    dataset: Union[Dataset, Dict, None],
    kwargs: Dict[str, Any],
    fields_to_exclude: Optional[List[str]] = None,
) -> Dict:
    # convert to dict to make transformations
    if isinstance(dataset, BaseDataset):
        dataset_dict = dataset.dict(exclude_none=True)
    elif dataset is not None and isinstance(dataset, dict):
        dataset_dict = dataset
    else:
        dataset_dict = kwargs

    if fields_to_exclude:
        for field in fields_to_exclude:
            if field in dataset_dict:
                del dataset_dict[field]

    return dataset_dict


def normalize_data(
    data: Union[str, dict, List[List[Any]], "pd.DataFrame", "gpd.GeoDataFrame"]
) -> Union[str, dict, List[List[Any]]]:

    if pd is not None and isinstance(data, pd.DataFrame):
        # convert to CSV string
        return data.to_csv()
    elif gpd is not None and isinstance(data, gpd.GeoDataFrame):
        # convert to GeoJSON FeatureCollection
        return data.to_json()
    else:
        return data


###########
# METHODS #
###########


class BaseDatasetApiMethods:
    @abstractmethod
    def add_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Optional[Dataset]:
        ...

    @abstractmethod
    def add_tile_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Optional[Dataset]:
        ...

    @abstractmethod
    def update_dataset(
        self, dataset_id: str, values: Union[Dataset, Dict, None] = None, **kwargs: Any
    ) -> Optional[Dataset]:
        ...

    @abstractmethod
    def remove_dataset(self, dataset_id: str) -> None:
        ...

    @abstractmethod
    def replace_dataset(
        self, this_dataset_id: str, with_dataset: Optional[dict] = None, **kwargs: Any
    ) -> Optional[Dataset]:
        ...


class BaseInteractiveDatasetApiMethods:
    @abstractmethod
    def get_datasets(self) -> List[Dataset]:
        ...

    @abstractmethod
    def get_dataset_by_id(self, dataset_id: str) -> Optional[Dataset]:
        ...

    @abstractmethod
    def get_dataset_with_data(self, dataset_id: str) -> Optional[DatasetWithData]:
        ...


class DatasetApiNonInteractiveMixin(BaseDatasetApiMethods):
    """Dataset methods that are supported in non-interactive (i.e. pure HTML) maps"""

    transport: BaseNonInteractiveTransport

    @validate_kwargs(positional_only=["dataset"])
    def add_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> None:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
                    metadata: (Dict): Object containing tileset metadata (for tiled datasets).

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
            metadata: (Dict): Object containing tileset metadata (for tiled datasets).
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields"])

        if "data" in dataset:
            dataset["data"] = normalize_data(dataset["data"])

        action = AddDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["dataset"])
    def add_tile_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> None:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    metadata: (Dict): Object containing tileset metadata.

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            metadata: (Dict): Object containing tileset metadata.
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields", "data"])

        action = AddTileDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["dataset_id", "values"])
    def update_dataset(
        self,
        dataset_id: str,
        values: Union[_DatasetUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> None:
        """Updates an existing dataset with given values.

        Args:
            dataset_id: The identifier of the dataset to update.
            values: The values to update.

        Kwargs:
            label: Displayable dataset label.
            color: Color label of the dataset.

        Returns:
            Updated dataset record.
        """

        action = UpdateDatasetAction(
            dataset_id=dataset_id, values=values if values else kwargs
        )
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["dataset_id"])
    def remove_dataset(self, dataset_id: str) -> None:
        """Removes a dataset from the map.

        Args:
            dataset_id: The identifier of the dataset to remove

        Returns:
            None
        """
        action = RemoveDatasetAction(dataset_id=dataset_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["this_dataset_id", "with_dataset"])
    def replace_dataset(
        self,
        this_dataset_id: str,
        with_dataset: Union[Dataset, Dict, None] = None,
        **kwargs: Any,
    ) -> None:
        """Replaces a given dataset with a new one.

        Args:
            this_dataset_id: Identifier of the dataset to replace.
            with_dataset: Dataset details to replace the dataset with.

        Returns:
            The dataset that's now in use.
        """
        with_dataset = normalize_dataset(with_dataset, kwargs, ["fields", "data"])

        action = ReplaceDatasetAction(
            this_dataset_id=this_dataset_id,
            with_dataset=with_dataset,
        )
        self.transport.send_action(action=action, response_class=None)


class DatasetApiInteractiveMixin(
    BaseDatasetApiMethods, BaseInteractiveDatasetApiMethods
):

    transport: BaseInteractiveTransport

    def get_datasets(self) -> List[Dataset]:
        """Gets all the datasets currently available in the map.

        Returns:
            An array of datasets.
        """
        action = GetDatasetsAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[Dataset]
        )

    @validate_kwargs(positional_only=["dataset_id"])
    def get_dataset_by_id(self, dataset_id: str) -> Optional[Dataset]:
        """Retrieves a dataset by its identifier if it exists.

        Args:
            dataset_id: Identifier of the dataset to get.

        Returns:
            Dataset with a given identifier, or null if one doesn't exist.
        """
        action = GetDatasetByIdAction(dataset_id=dataset_id)
        return self.transport.send_action(action=action, response_class=Dataset)  # type: ignore

    @validate_kwargs(positional_only=["dataset"])
    def add_dataset(
        self,
        dataset: Optional[dict] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Dataset:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
                    metadata: (Dict): Object containing tileset metadata (for tiled datasets).

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            data: (Union[str, dict, List[List[Any]]]): Data used to create a dataset, in CSV, JSON, or GeoJSON format (for local datasets).
            metadata: (Dict): Object containing tileset metadata (for tiled datasets).
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields"])

        if "data" in dataset:
            dataset["data"] = normalize_data(dataset["data"])

        action = AddDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset"])
    def add_tile_dataset(
        self,
        dataset: Union[Dataset, Dict, None] = None,
        *,
        auto_create_layers: bool = True,
        center_map: bool = True,
        **kwargs: Any,
    ) -> Dataset:
        """Adds a new dataset to the map.

        Args:
            dataset:
                A dictionary with the following format. Unused if these attributes are provided through kwargs.
                    id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
                    label (Optional[str]): Displayable dataset label.
                    color: (Optional[Tuple[float, float, float]]): Color label of the dataset
                    metadata: (Dict): Object containing tileset metadata.

        Kwargs:
            auto_create_layers:
                Whether to attempt and create new layer(s) when adding a dataset. Defaults to True.
            center_map:
                Whether to center the map to fit the new layer bounds. Defaults to True.
            id (Optional[str]): Unique identifier of the dataset. If not provided, a random id will be generated.
            label (Optional[str]): Displayable dataset label.
            color: (Optional[Tuple[float, float, float]]): Color label of the dataset
            metadata: (Dict): Object containing tileset metadata.
        """

        dataset = normalize_dataset(dataset, kwargs, ["fields", "data"])

        action = AddTileDatasetAction(
            dataset=dataset,
            auto_create_layers=auto_create_layers,
            center_map=center_map,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset_id", "values"])
    def update_dataset(
        self,
        dataset_id: str,
        values: Union[_DatasetUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> Dataset:
        """Updates an existing dataset with given values.

        Args:
            dataset_id: The identifier of the dataset to update.
            values: The values to update.

        Kwargs:
            label: Displayable dataset label.
            color: Color label of the dataset.

        Returns:
            Updated dataset record.
        """

        action = UpdateDatasetAction(
            dataset_id=dataset_id, values=values if values else kwargs
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset_id"])
    def remove_dataset(self, dataset_id: str) -> None:
        """Removes a dataset from the map.

        Args:
            dataset_id: The identifier of the dataset to remove

        Returns:
            None
        """
        action = RemoveDatasetAction(dataset_id=dataset_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["this_dataset_id", "with_dataset"])
    def replace_dataset(
        self,
        this_dataset_id: str,
        with_dataset: Union[Dataset, Dict, None] = None,
        **kwargs: Any,
    ) -> Dataset:
        """Replaces a given dataset with a new one.

        Args:
            this_dataset_id: Identifier of the dataset to replace.
            with_dataset: Dataset details to replace the dataset with.

        Returns:
            The dataset that's now in use.
        """
        with_dataset = normalize_dataset(with_dataset, kwargs, ["fields", "data"])

        action = ReplaceDatasetAction(
            this_dataset_id=this_dataset_id,
            with_dataset=with_dataset,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=Dataset  # type: ignore
        )

    @validate_kwargs(positional_only=["dataset_id"])
    def get_dataset_with_data(self, dataset_id: str) -> Optional[DatasetWithData]:
        """Retrieves a dataset record with its data for a given dataset if it exists.

        Args:
            dataset_id: Identifier of the dataset to get the record for.

        Returns:
            Dataset record with its data based on an identifier, or None if one doesn't exist.
        """
        action = GetDatasetWithDataAction(dataset_id=dataset_id)
        return self.transport.send_action(action=action, response_class=DatasetWithData)



from unittest.mock import patch

from unfolded.map_sdk.map.html import HTMLMap

from .fixtures import EARTHQUAKES_DF


class TestHTMLMap:
    @patch("unfolded.map_sdk.transport.html.__version__", "test")
    @patch("unfolded.map_sdk.transport.html.uuid4", lambda: "test-div")
    def test_template_rendering(self, snapshot):

        m = HTMLMap()
        m.add_dataset(data=EARTHQUAKES_DF, label="Earthquakes")
        m.set_view(zoom=7)
        snapshot.assert_match(m._repr_html_(), "map.html")

    def test_html_map_action_list(self):

        m = HTMLMap()
        transport = m.transport

        m.set_view(zoom=7)
        m.set_view_limits(min_zoom=5)

        assert transport.serialized_actions == [
            {"args": [{"zoom": 7}], "options": {"index": 0}, "funcName": "setView"},
            {
                "args": [{"minZoom": 5}],
                "options": {"index": 0},
                "funcName": "setViewLimits",
            },
        ]

        m.set_theme(preset="light")

        # action_list remembers all previous actions
        assert transport.serialized_actions == [
            {"args": [{"zoom": 7}], "options": {"index": 0}, "funcName": "setView"},
            {
                "args": [{"minZoom": 5}],
                "options": {"index": 0},
                "funcName": "setViewLimits",
            },
            {"args": [{"preset": "light", "options": {}}], "funcName": "setUiTheme"},
        ]



import sys
from typing import Any, Union, overload

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from unfolded.map_sdk.environment import CURRENT_ENVIRONMENT, Environment
from unfolded.map_sdk.errors import MapSDKException
from unfolded.map_sdk.map import HTMLMap, SyncWidgetMap

__all__ = ("create_map",)

DATABRICKS_HTML_MAP_MSG = """Databricks environment detected, using HTML renderer.
SDK function calls after a map is rendered will not update the map automatically,
the map must be re-rendered to update.
To hide this warning in the future, pass `renderer="html"`.
"""

# Note: not possible (I think) to have this typing overload understand when the current environment
# is databricks
@overload
def create_map(*, renderer: Literal["html"], **kwargs: Any) -> HTMLMap:
    ...


@overload
def create_map(
    *, renderer: Literal["widget", None] = None, **kwargs: Any
) -> SyncWidgetMap:
    ...


def create_map(
    *, renderer: Literal["html", "widget", None] = None, **kwargs: Any
) -> Union[HTMLMap, SyncWidgetMap]:
    """Create an unfolded map

    Args:
        renderer (str): Which renderer to use for the map, either "html" or "widget".
                        Default: "widget" if supported by your environment.
    """
    if CURRENT_ENVIRONMENT == Environment.DATABRICKS:
        if renderer == "widget":
            raise MapSDKException(
                "Cannot use widget renderer in Databricks environment"
            )
        elif renderer is None:
            sys.stderr.write(DATABRICKS_HTML_MAP_MSG)
        return HTMLMap(**kwargs)
    elif renderer == "html":
        return HTMLMap(**kwargs)
    else:
        return SyncWidgetMap(**kwargs)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import os
from glob import glob
from os.path import join as pjoin

from jupyter_packaging import (
    combine_commands,
    create_cmdclass,
    ensure_targets,
    get_version,
    install_npm,
)
from setuptools import find_namespace_packages, setup

HERE = os.path.dirname(os.path.abspath(__file__))

with open("README.md") as f:
    readme = f.read()

# The name of the project
name = "unfolded.map-sdk"
# Directory as a tuple since we use namespace packaging
code_directory = ("unfolded", "map_sdk")

# Get the version
version = get_version(pjoin(*code_directory, "_version.py"))

# Representative files that should exist after a successful build
jstargets = [
    pjoin(HERE, *code_directory, "nbextension", "index.js"),
    pjoin(HERE, "lib", "plugin.js"),
]

package_data_spec = {name: ["nbextension/**js*", "labextension/**"]}

data_files_spec = [
    (
        "share/jupyter/nbextensions/unfolded/map_sdk",
        pjoin(*code_directory, "nbextension"),
        "**",
    ),
    # These are intended to be npm package name
    (
        "share/jupyter/labextensions/@unfolded/jupyter-map-sdk",
        pjoin(*code_directory, "labextension"),
        "**",
    ),
    ("share/jupyter/labextensions/@unfolded/jupyter-map-sdk", ".", "install.json"),
    ("etc/jupyter/nbconfig/notebook.d", ".", "unfolded.map_sdk.json"),
]

cmdclass = create_cmdclass(
    "jsdeps", package_data_spec=package_data_spec, data_files_spec=data_files_spec
)

cmdclass["jsdeps"] = combine_commands(
    install_npm(HERE, build_cmd="build:prod", npm=["yarn"]),
    ensure_targets(jstargets),
)

setup_args = dict(
    name=name,
    description="Jupyter Widget for Unfolded.ai Maps",
    long_description=readme,
    long_description_content_type="text/markdown",
    version=version,
    scripts=glob(pjoin("scripts", "*")),
    cmdclass=cmdclass,
    packages=find_namespace_packages(include=["unfolded.*"]),
    author="UnfoldedInc.",
    author_email="info@unfolded.ai",
    license="(c) 2022 Foursquare Labs, Inc. Terms available at https://www.unfolded.ai/terms-of-service/",
    platforms="Linux, Mac OS X, Windows",
    keywords=["Jupyter", "Widgets", "IPython"],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: Jupyter",
    ],
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "ipykernel>=4,<7",
        "ipywidgets>=7.0.0,<8",
        "jupyter-ui-poll>=0.2.1,<0.3.0",
        "pydantic>=1.9,<2.0",
        "geojson-pydantic>=0.4.2,<0.5.0",
        'typing_extensions;python_version<"3.8"',
    ],
    extras_require={
        "test": [
            "pytest>=4.6",
            "pytest-cov",
            "pytest-snapshot",
            "pytest-mock",
            "nbval",
            "mypy==0.961",
            "jupyter_packaging",
            # This is the version used in ipywidgets ui tests https://github.com/jupyter-widgets/ipywidgets/blob/569ae06e53242c7f78790caf29a7e63ba1a06133/.github/workflows/tests.yml#L167
            "jupyterlab~=3.2",
            "pandas",
            "jupytext==1.13.6",
        ],
        "geopandas": ["geopandas>=0.7.0", "pandas"],
        "examples": [
            # Any requirements for the examples to run
        ],
        "docs": [
            "jupyter_sphinx",
            "nbsphinx",
            "nbsphinx-link",
            "pytest_check_links",
            "pypandoc",
            "recommonmark",
            "sphinx>=1.5",
            "sphinx_rtd_theme",
        ],
    },
    entry_points={},
)

if __name__ == "__main__":
    setup(**setup_args)



from pathlib import Path
from typing import Dict, Optional

from unfolded.map_sdk.map.base import (
    DEFAULT_MAP_STYLE,
    BasemapParams,
    BaseNonInteractiveMap,
)
from unfolded.map_sdk.transport.html import HTMLTransport

TEMPLATE_DIR = (Path(__file__).parent / ".." / "templates").resolve()


class HTMLMap(BaseNonInteractiveMap):
    transport: HTMLTransport

    def __init__(
        self,
        style: Optional[Dict] = None,
        basemaps: Optional[Dict] = None,
    ):

        if style:
            self.style = {**DEFAULT_MAP_STYLE, **style}
        else:
            self.style = DEFAULT_MAP_STYLE

        if basemaps:
            validated_basemaps = BasemapParams(**basemaps)
            self.basemaps = validated_basemaps.dict(by_alias=True, exclude_none=True)
        else:
            self.basemaps = {}

        self.transport = HTMLTransport()
        self.rendered = False

    def _repr_html_(self) -> str:
        return self.render()

    def render(self) -> str:
        return self.transport.render_template(style=self.style, basemaps=self.basemaps)



from pytest_mock import MockerFixture

from unfolded.map_sdk.map.widget import SyncWidgetMap


class TestAddDataset:
    def test_raster_custom_stac_item(
        self, mocker: MockerFixture, sentinel2_stac_item: dict
    ):
        mocked_send_widget_message = mocker.patch(
            "unfolded.map_sdk.transport.widget.BlockingWidgetTransport._send_widget_message"
        )
        _ = mocker.patch(
            "unfolded.map_sdk.transport.widget.BlockingWidgetTransport._wait_for_future"
        )

        unfolded_map = SyncWidgetMap()

        raster_tile_dataset = {"type": "raster-tile", "metadata": sentinel2_stac_item}
        unfolded_map.add_dataset(raster_tile_dataset)

        # TODO: assert that the right object was sent to JS using call_args_list
        print(mocked_send_widget_message.call_args_list)

    def test_raster_custom_stac_collection(
        self, mocker: MockerFixture, sentinel2_stac_collection: dict
    ):
        mocked_send_widget_message = mocker.patch(
            "unfolded.map_sdk.transport.widget.BlockingWidgetTransport._send_widget_message"
        )
        _ = mocker.patch(
            "unfolded.map_sdk.transport.widget.BlockingWidgetTransport._wait_for_future"
        )

        unfolded_map = SyncWidgetMap()

        raster_tile_dataset = {
            "type": "raster-tile",
            "metadata": sentinel2_stac_collection,
        }
        unfolded_map.add_dataset(raster_tile_dataset)

        # TODO: assert that the right object was sent to JS using call_args_list
        print(mocked_send_widget_message.call_args_list)



from unfolded.map_sdk.api import map_api

DUMMY_MESSAGE_UUID = "ede3fafc-4945-41b7-ac64-16faa896a47a"
DATASET_UUID = "0e939627-8ea7-4db6-8c92-3dd943166a01"
LAYER_UUID = "997c21eb-604b-49df-add9-b2fcc615a243"
FILTER_UUID = "e02b67ea-20d4-4613-808e-9a8fdbd81e6f"

ACTION_FIXTURES = [
    (
        map_api.GetViewAction(index=0),
        {
            "type": "v1/map-sdk-get-view",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [0],
        },
    ),
    (
        map_api.SetViewAction(
            view={"latitude": 47.271057, "longitude": 8.650367, "zoom": 5},
            index=0,
        ),
        {
            "type": "v1/map-sdk-set-view",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                {
                    "longitude": 8.650367,
                    "latitude": 47.271057,
                    "zoom": 5.0,
                }
            ],
            "options": {"index": 0},
        },
    ),
    (
        map_api.GetViewLimitsAction(index=0),
        {
            "type": "v1/map-sdk-get-view-limits",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [0],
        },
    ),
    (
        map_api.SetViewLimitsAction(
            view_limits={"min_zoom": 3, "max_zoom": 8}, index=0
        ),
        {
            "type": "v1/map-sdk-set-view-limits",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                {
                    "minZoom": 3,
                    "maxZoom": 8,
                }
            ],
            "options": {"index": 0},
        },
    ),
    (
        map_api.GetMapControlVisibilityAction(),
        {
            "type": "v1/map-sdk-get-map-control-visibility",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
    (
        map_api.SetMapControlVisibilityAction(
            visibility={"legend": False, "toggle_3d": True}
        ),
        {
            "type": "v1/map-sdk-set-map-control-visibility",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                {
                    "legend": False,
                    "toggle-3d": True,
                }
            ],
        },
    ),
    (
        map_api.GetSplitModeAction(),
        {
            "type": "v1/map-sdk-get-split-mode",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
    (
        map_api.SetSplitModeAction(
            split_mode="swipe",
            options={
                "layers": [[LAYER_UUID], []],
                "is_view_synced": True,
                "is_zoom_synced": True,
            },
        ),
        {
            "type": "v1/map-sdk-set-split-mode",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                "swipe",
                {
                    "layers": [[LAYER_UUID], []],
                    "isViewSynced": True,
                    "isZoomSynced": True,
                },
            ],
        },
    ),
    (
        map_api.SetThemeAction(
            theme={"preset": "light", "options": {"background_color": "blue"}}
        ),
        {
            "type": "v1/map-sdk-set-theme",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [{"preset": "light", "options": {"backgroundColor": "blue"}}],
        },
    ),
    (
        map_api.GetMapConfigAction(),
        {
            "type": "v1/map-sdk-get-map-config",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
    (
        map_api.SetMapConfigAction(
            config={"map": "config"},
            options={
                "additional_datasets": [
                    {"id": "dataset-id", "type": "local", "data": "dataset-data"}
                ]
            },
        ),
        {
            "type": "v1/map-sdk-set-map-config",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                {"map": "config"},
                {
                    "additionalDatasets": [
                        {"id": "dataset-id", "type": "local", "data": "dataset-data"}
                    ]
                },
            ],
        },
    ),
    (
        map_api.GetMapStylesAction(),
        {
            "type": "v1/map-sdk-get-map-styles",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
]
